import json
import ckan.plugins.toolkit as toolkit
import ckan.model
import pylons
import dateutil.parser

from budgetdatapackage import BudgetDataPackage, BudgetResource

import logging
log = logging.getLogger(__name__)

class BudgetDataPackageController(toolkit.BaseController):

    def descriptor(self, id, resource_id):
        # Set context
        context = {'model': ckan.model, 'session': ckan.model.Session,
                   'user': pylons.c.user or pylons.c.author}

        # Get resource
        resource = toolkit.get_action('resource_show')(
            context, {'id': resource_id})

        if not resource.get('BudgetDataPackage', False):
            raise toolkit.ObjectNotFound(
                'No descriptor file for this resource')

        # If last modified hasn't been set we set it as time of creation
        last_modified = resource.get('last_modified')
        if last_modified is None:
            last_modified = resource['created']

        # Create the budget data package resource
        bdgt_resource = BudgetResource(
            name=resource['id'],
            schema=resource['schema'],
            url=resource['url'],
            currency=resource['currency'],
            fiscalYear=resource['year'],
            granularity=resource['granularity'],
            type=resource['type'],
            status=resource['status'],
            datePublished=dateutil.parser.parse(
                resource['created']).date().isoformat(),
            dateLastUpdated=dateutil.parser.parse(
                last_modified).date().isoformat(),
            location=resource['country'])

        # datapackage_uri and is_local are added but not needed
        # so we clean up our budget resource
        del bdgt_resource['datapackage_uri']
        del bdgt_resource['is_local']

        # Create the budget data package
        bdpkg = BudgetDataPackage(
            name=id,
            title=resource['name'],
            description=resource['description'],
            resources=[bdgt_resource]
        )

        # Return the budget data package descriptor (json)
        toolkit.response.headers['Content-Type'] = 'application/json'
        return json.dumps(bdpkg)



import ckan.plugins as plugins
from ckan.logic.validators import int_validator

import datetime
import os.path
import json

from ckanext.budgets.lib.budgetdatapackage import BudgetDataPackage
from ckanext.budgets import exceptions

import logging
log = logging.getLogger(__name__)



class BudgetDataPackagePlugin(plugins.SingletonPlugin,
                              plugins.toolkit.DefaultDatasetForm):
    """Budget Data Package creator

    The plugin hooks into file upload/import and analyses uploaded/imported
    resources to see if they might be a budget data package. If a resource
    looks like a budget data package resource, the plugin allows the user to
    create a budget data package descriptor file (datapackage.json)
    """

    plugins.implements(plugins.IConfigurable)
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IDatasetForm)
    plugins.implements(plugins.IResourceController)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IRoutes, inherit=True)

    def __init__(self, *args, **kwargs):
        self.countries = {}
        self.currencies = {}
        self.statuses = {}
        self.data = None

    def configure(self, config):
        """
        Initialize the plugin. This creates a data object which holds a
        BudgetDataPackage parser which operates based on a specification
        which is either provided in the config via:
        ``ckan.budgets.specification`` or the included version.
        """

        specification = config.get(
            'ckan.budgets.specification',
            os.path.join(os.path.dirname(__file__),
                         'data', 'bdp', 'schema.json'))
        self.data = BudgetDataPackage(specification)

        countries_json = config.get(
            'ckan.budgets.countries',
            os.path.join(os.path.dirname(__file__),
                         'data', 'countries.json'))
        with open(countries_json) as country_list:
            self.countries = json.load(country_list)

        country = config.get('ckan.budgets.default.country', None)
        if country is not None:
            self.default_country = country.upper()
            if self.default_country not in self.countries:
                raise ValueError('Uknown country code "{code}"'.format(
                    code=country))
        else:
            self.default_currency = None

        currencies_json = config.get(
            'ckan.budget.currencies',
            os.path.join(os.path.dirname(__file__),
                         'data', 'currencies.json'))
        with open(currencies_json) as currency_list:
            self.currencies = json.load(currency_list)

        currency = config.get('ckan.budgets.default.currency', None)
        if currency is not None:
            self.default_currency = currency.upper()
            if self.default_currency not in self.currencies:
                raise ValueError('Unknown currency code "{code}"'.format(
                    code=currency))
        else:
            self.default_currency = None

        statuses_json = config.get(
            'ckan.budget.statuses',
            os.path.join(os.path.dirname(__file__),
                         'data', 'bdp', 'statuses.json'))
        with open(statuses_json) as statuses_list:
            self.statuses = json.load(statuses_list)

    def update_config(self, config):
        pass
        plugins.toolkit.add_template_directory(
            config,
            os.path.join(os.path.dirname(__file__), 'form', 'templates'))


    def before_map(self, map_):
        controller = "ckanext.budgets.controllers:BudgetDataPackageController"
        map_.connect(
            "budgetdatapackage_descriptor",
            "/dataset/{id}/resource/{resource_id}/datapackage.json",
            controller=controller,
            action="descriptor")

        return map_

    def _sort_value(self, dictionary):
        return sorted(dictionary.iteritems(), key=lambda (key, value): value)

    def get_countries(self):
        countries = [{'value':code, 'text':name}
                     for (code, name) in self._sort_value(self.countries)]
        countries.insert(0, {'value':'', 'text':'-- Not applicable --'})
        return countries

    def get_country(self, default):
        if not self.default_country:
            return default
        return self.default_country

    def get_currencies(self):
        currencies = [{'value':code, 'text':name}
                      for (code, name) in self._sort_value(self.currencies)]
        currencies.insert(0, {'value':'', 'text':'-- Not applicable --'})
        return currencies

    def get_currency(self, default):
        if not self.default_currency:
            return default
        return self.default_currency

    def get_statuses(self):
        statuses = [{'value':status_obj['status'], 'text':status_obj['label']}
                    for status_obj  in self.statuses]
        statuses.insert(0, {'value':'', 'text':'-- Not applicable --'})
        return statuses

    def get_helpers(self):
        return {
            'get_countries': self.get_countries,
            'get_country': self.get_country,
            'get_currencies': self.get_currencies,
            'get_currency': self.get_currency,
            'get_statuses': self.get_statuses
        }

    def country_validator(self, value, context):
        if value and value not in self.countries:
            raise plugins.toolkit.Invalid(
                'Country value ({0}) is invalid'.format(value))
        return value

    def currency_validator(self, value, context):
        if value and value not in self.currencies:
            raise plugins.toolkit.Invalid(
                'Currency value ({0}) is invalid'.format(value))
        return value

    def year_validator(self, value, context):
        value = int_validator(value, context)
        if value and value < 0:
            raise plugins.toolkit.Invalid(
                'Year value ({0}) is invalid'.format(value))
        return str(value)

    def status_validator(self, value, context):
        if value and value not in [s['status'] for s in self.statuses]:
            raise plugins.toolkit.Invalid(
                'Status value ({0}) is invalid'.format(value))
        return value

    def schema_validator(self, value, context):
        if value:
            return value

        return []

    @property
    def resource_schema_additions(self):
        return {
            'country': [
                self.country_validator
            ],
            'currency': [
                self.currency_validator
            ],
            'year': [
                self.year_validator
            ],
            'status': [
                self.status_validator
            ],
            'schema': [
                self.schema_validator
            ]
        }

    def create_package_schema(self):
        schema = super(BudgetDataPackagePlugin, self).create_package_schema()
        schema['resources'].update(self.resource_schema_additions)
        return schema

    def update_package_schema(self):
        schema = super(BudgetDataPackagePlugin, self).update_package_schema()
        schema['resources'].update(self.resource_schema_additions)
        return schema

    def show_package_schema(self):
        schema = super(BudgetDataPackagePlugin, self).show_package_schema()
        schema['resources'].update(self.resource_schema_additions)
        return schema

    def is_fallback(self):
        return True

    def package_types(self):
        return ['dataset']

    def in_resource(self, field, resource):
        """
        Return True if resource contains a valid value for the field
        (not an empty or None value)
        """
        resource_field = resource.get(field, None)
        return resource_field is not None and resource_field != ''

    def are_budget_data_package_fields_filled_in(self, resource):
        """
        Check if the budget data package fields are all filled in because
        if not then this can't be a budget data package
        """
        fields = ['country', 'currency', 'year', 'status']
        return all([self.in_resource(f, resource) for f in fields])

    def generate_budget_data_package(self, resource):
        """
        Try to grab a budget data package schema from the resource.
        The schema only allows fields which are defined in the budget
        data package specification. If a field is found that is not in
        the specification this will return a NotABudgetDataPackageException
        and in that case we can just return and ignore the resource
        """

        # Return if the budget data package fields have not been filled in
        if not self.are_budget_data_package_fields_filled_in(resource):
            return

        try:
            resource['schema'] = self.data.schema
        except exceptions.NotABudgetDataPackageException:
            log.debug('Resource is not a Budget Data Package')
            resource['schema'] = []
            return

        # If the schema fits, this can be exported as a budget data package
        # so we add the missing metadata fields to the resource.
        resource['BudgetDataPackage'] = True
        resource['standard'] = self.data.version
        resource['granularity'] = self.data.granularity
        resource['type'] = self.data.budget_type

    def before_create(self, context, resource):
        """
        When triggered the resource which can either be uploaded or linked
        to will be parsed and analysed to see if it possibly is a budget
        data package resource (checking if all required headers and any of
        the recommended headers exist in the csv).

        The budget data package specific fields are then appended to the
        resource which makes it useful for export the dataset as a budget
        data package.
        """

        # If the resource is being uploaded we load the uploaded file
        # If not we load the provided url
        if resource.get('upload', '') == '':
            self.data.load(resource['url'])
        else:
            self.data.load(resource['upload'].file)

        self.generate_budget_data_package(resource)

    def after_create(self, context, resource):
        pass

    def before_update(self, context, current, resource):
        """
        If the resource has changed we try to generate a budget data
        package, but if it hasn't then we don't do anything
        """

        # Return if the budget data package fields have not been filled in
        if not self.are_budget_data_package_fields_filled_in(resource):
            return

        if resource.get('upload', '') == '':
            # If it isn't an upload we check if it's the same url
            if current['url'] == resource['url']:
                # Return if it's the same
                return
            else:
                self.data.load(resource['url'])
        else:
            self.data.load(resource['upload'].file)

        self.generate_budget_data_package(resource)

    def after_update(self, context, resource):
        pass

    def before_delete(self, context, resource, resources):
        pass

    def after_delete(self, context, resources):
        pass

    def before_show(self, resource):
        return resource



class NotABudgetDataPackageException(Exception):
    '''
    The exception type that is raised when a resource file does not conform
    with the Budget Data Package specification.
    '''
    pass

class ResourceNotLoadedException(Exception):
    """
    The exception type that is raised when methods that rely on a resource
    being loaded are called, before the resource has been loaded.
    """
    pass



from setuptools import setup, find_packages
import sys, os

version = '0.2.0'

description = "CKAN extension for creating/distributing budget data packages"
with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name='ckanext-budgets',
    version=version,
    description=description,
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='ckan',
    author='Tryggvi Bjorgvinsson',
    author_email='tryggvi.bjorgvinsson@okfn.org',
    url='https://github.com/tryggvib/ckanext-budgets',
    license='AGPLv3',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=['ckanext', 'ckanext.budgets'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # -*- Extra requirements: -*-
        'unicodecsv',
	'requests',
	'six',
        'py-dateutil',
        'budgetdatapackage'
    ],
    entry_points='''
        [ckan.plugins]
        budgets=ckanext.budgets.plugin:BudgetDataPackagePlugin
    ''',
)



import requests
import json
import six
import cgi
import cStringIO
import unicodecsv as csv
from contextlib import closing

from ckanext.budgets import exceptions

class BudgetDataPackage(object):
    """
    Budget data is stored in CSV files. This class wraps a CSV file to limit
    access to the file by providing headers and other helpful methods.
    """

    def __init__(self, specification, *args, **kwargs):
        """
        Initialialize the BudgetDataPackage object with the specification
        that should be used for processing budget data package resources
        """
        self.headers = None
        self.specification = self._get_specification(specification)

    def _get_specification(self, specification):
        """
        Read the specification provided. It can either be a url or a file
        location.
        """
        result = six.moves.urllib.parse.urlparse(specification)

        # If the specification has an http or an https scheme we can
        # retrieve it via an HTTP get request, else we try to open it
        # as a file.
        if result.scheme in ['http', 'https']:
            response = requests.get(specification)
            spec_json = response.json()
        else:
            with open(specification, 'r') as spec_file:
                spec_json = json.load(spec_file)

        return spec_json

    def _get_headers(self, resource):
        """
        Get CSV file headers from the provided resource.
        """

        # If the resource is a file we just open it up with the csv
        # reader (after being sure we're reading from the beginning
        # of the file
        if type(resource) == file:
            resource.seek(0)
            reader = csv.reader(resource)
        # If the resource is a basestring it is either a url or a file
        # location, so similarly to the specification mechanism we either
        # access it with an HTTP get request or by opening the file.
        elif isinstance(resource, basestring):
            result = six.moves.urllib.parse.urlparse(resource)
            if result.scheme in ['http', 'https']:
                with closing(requests.get(resource, stream=True)) as response:
                    # Headers are alway the first row of a CSV file
                    # so it's enought to just get the first line and
                    # hopefully save bandwidth
                    header_row = response.iter_lines().next()
            else:
                # It may seem weird to open up a csv file, read its header row
                # and then StringIO that into a new csv reader but this file
                # we want to close and we want the same interface for all
                with open(resource) as resource_file:
                    reader = csv.reader(resource_file)
                    header_row = reader.next()

            reader = csv.reader(cStringIO.StringIO(header_row))
        else:
            raise IOError('Resource type not supported')
        return reader.next()    

    def load(self, resource):
        """
        Load a resource into the BudgetDataPackage parser. The resource
        can either be an open file object, a url or a file path.
        """
        self.headers = self._get_headers(resource)

    @property
    def version(self):
        """
        Version of budget data package being used, according to the
        specification
        """
        return self.specification.get('version', None)

    @property
    def schema(self):
        """
        The generated budget data package schema for this resource.
        If the resource has any fields that do not conform to the
        provided specification this will raise a
        NotABudgetDataPackageException.
        """

        if self.headers is None:
            raise exceptions.NoResourceLoadedException(
                'Resource must be loaded to find schema')
        try:
            fields = self.specification.get('fields', {})
            parsed = {
                'primaryKey': 'id',
                'fields': [{
                        'name': header,
                        'type': fields[header]['type'],
                        'description': fields[header]['description']
                        } for header in self.headers]
                }
        except KeyError:
            raise exceptions.NotABudgetDataPackageException(
                'Includes other fields than the Budget Data Package fields')
        return parsed

    @property
    def granularity(self):
        """
        Granularity of the budget resource, either aggregated or
        transactional.
        """

        if self.headers is None:
            raise exceptions.NoResourceLoadedException(
                'Resource must be loaded to find granularity')

        return 'transactional' if 'date' in self.headers else 'aggregated'

    @property
    def budget_type(self):
        """
        Type of the budget resource, either expenditure or revenue.
        """

        if self.headers is None:
            raise exceptions.NoResourceLoadedException(
                'Resource must be loaded to find type')

        return 'expenditure' if 'admin' in self.headers else 'revenue'
